library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- DATA_REG (slv_reg0):
--    ---------------------------------------------------------------------------------
--    | 31 | ....................................................................|  0 | 
--    ---------------------------------------------------------------------------------
--    ---------------------------------------------------------------------------------
-- RD:|                             data_to_read (from FIFO)                          | 
--    ---------------------------------------------------------------------------------
-- WR:|                             data_to_send (to FIFO)                            | 
--    ---------------------------------------------------------------------------------

-- STATUS_CONTROL_REG (slv_reg1):
--    ---------------------------------------------------------------------------------
--    | 31 | 30 | 29 | 28 | 27 | 26 | 25 | 24 | 23 | 22 | 21 | 20 | 19 | 18 | 17 | 16 | 
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- RD:| HW |  HW| HW | HW |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |
--    |VER3|VER2|VER1|VER0|    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- WR:|  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    ---------------------------------------------------------------------------------
--    | 15 | 14 | 13 | 12 | 11 | 10 |  9 |  8 |  7 |  6 |  5 |  4 |  3 |  2 |  1 |  0 |  
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    | rx | rx | tx | tx | tx | rx |    |    |    |rx=0|    |
-- RD:|  0 |  0 |  0 |  0 |  0 |fifo|fifo|fifo|fifo|done|done|  0 |  0 |  0 |tx=1|busy|
--    |    |    |    |    |    |empt|full|empt|full|    |    |    |    |    |mode|    |
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    | rx |strt|strt| rx | tx |
-- WR:|  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |fifo|next|next|mode|mode|
--    |    |    |    |    |    |    |    |    |    |    |    | rd | tx | rx | set| set|
--    ---------------------------------------------------------------------------------

-- CONFIG_REG (slv_reg2):
--    ---------------------------------------------------------------------------------
--    | 31 | 30 | 29 | 28 | 27 | 26 | 25 | 24 | 23 | 22 | 21 | 20 | 19 | 18 | 17 | 16 | 
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- RD:|  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- WR:|  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    ---------------------------------------------------------------------------------
--    | 15 | 14 | 13 | 12 | 11 | 10 |  9 |  8 |  7 |  6 |  5 |  4 |  3 |  2 |  1 |  0 |  
--    ---------------------------------------------------------------------------------
--    | cfg|                                  |                                       |
-- RD:|busy|     config addr CC1000           |          config data CC1000           |
--    | CC |                                  |                                       |
--    ---------------------------------------------------------------------------------
--    |rd=0|                                  |                                       |
-- WR:|wr=1|     config addr CC1000           |          config data CC1000           |
--    |cfg |                                  |                                       |
--    ---------------------------------------------------------------------------------

-- ADC_REG (slv_reg3):
--    ---------------------------------------------------------------------------------
--    | 31 | 30 | 29 | 28 | 27 | 26 | 25 | 24 | 23 | 22 | 21 | 20 | 19 | 18 | 17 | 16 | 
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- RD:|  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  0 |
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
-- WR:|  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
--    ---------------------------------------------------------------------------------
--    ---------------------------------------------------------------------------------
--    | 15 | 14 | 13 | 12 | 11 | 10 |  9 |  8 |  7 |  6 |  5 |  4 |  3 |  2 |  1 |  0 |  
--    ---------------------------------------------------------------------------------
--    |ADC |data|    |    |                                                           |
-- RD:|busy|not |  0 |  0 |                        ADC data                           |
--    |    |used|    |    |                                                           |
--    ---------------------------------------------------------------------------------
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | ADC|
-- WR:|  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |  x |star|
--    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |conv|
--    ---------------------------------------------------------------------------------


entity cc1000_v3_0_S00_AXI is
	generic (
		-- Users to add parameters here
        DBG_LEN        :    integer :=128;    
        DATA_BYTES_MAX : integer                   := 64;  --max no. of data bytes; 4*n, where n=1,2,...
        PREAMBLE_BITS  : integer                   := 96;  --no. of preamble bits for tx,  must be EVEN
        SFD_BYTES      : std_logic_vector(0 to 15) := X"E39B";  -- Start of Frame Delimiter  
        CLK_DIV_ADC    : integer := 64; -- Clock of ADC CLK_ACD(<800kHz) = CLK/CLK_DIV_ADC
        CLK_DIV_CC1000 : integer := 3;  -- Config clock PCLK(<10MHz) of CC1000 f(PCLK) = CLK/(2*CLK_DIV_CC1000)                
		-- User parameters ends
		-- Do not modify the parameters beyond this line

		-- Width of S_AXI data bus
		C_S_AXI_DATA_WIDTH	: integer	:= 32;
		-- Width of S_AXI address bus
		C_S_AXI_ADDR_WIDTH	: integer	:= 4
	);
	port (
		-- Users to add ports here
        dbg_o            : out std_logic_vector(DBG_LEN-1 downto 0);    
        CC1000_DCLK_i       : in    std_logic;    -- to CC1000
        CC1000_DIO_io       : inout std_logic;    -- to CC1000
        CC1000_PCLK_o       : out   std_logic;    -- to CC1000
        CC1000_PDATA_io     : inout std_logic;    -- to CC1000
        CC1000_PALE_o       : out   std_logic;    -- to CC1000
        
        MCP3201_clk_o  : out   std_logic;
        MCP3201_data_i : in    std_logic;
        MCP3201_cs_o   : out   std_logic;
        
		-- User ports ends
		
		-- Do not modify the ports beyond this line

		-- Global Clock Signal
		S_AXI_ACLK	: in std_logic;
		-- Global Reset Signal. This Signal is Active LOW
		S_AXI_ARESETN	: in std_logic;
		-- Write address (issued by master, acceped by Slave)
		S_AXI_AWADDR	: in std_logic_vector(C_S_AXI_ADDR_WIDTH-1 downto 0);
		-- Write channel Protection type. This signal indicates the
    		-- privilege and security level of the transaction, and whether
    		-- the transaction is a data access or an instruction access.
		S_AXI_AWPROT	: in std_logic_vector(2 downto 0);
		-- Write address valid. This signal indicates that the master signaling
    		-- valid write address and control information.
		S_AXI_AWVALID	: in std_logic;
		-- Write address ready. This signal indicates that the slave is ready
    		-- to accept an address and associated control signals.
		S_AXI_AWREADY	: out std_logic;
		-- Write data (issued by master, acceped by Slave) 
		S_AXI_WDATA	: in std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
		-- Write strobes. This signal indicates which byte lanes hold
    		-- valid data. There is one write strobe bit for each eight
    		-- bits of the write data bus.    
		S_AXI_WSTRB	: in std_logic_vector((C_S_AXI_DATA_WIDTH/8)-1 downto 0);
		-- Write valid. This signal indicates that valid write
    		-- data and strobes are available.
		S_AXI_WVALID	: in std_logic;
		-- Write ready. This signal indicates that the slave
    		-- can accept the write data.
		S_AXI_WREADY	: out std_logic;
		-- Write response. This signal indicates the status
    		-- of the write transaction.
		S_AXI_BRESP	: out std_logic_vector(1 downto 0);
		-- Write response valid. This signal indicates that the channel
    		-- is signaling a valid write response.
		S_AXI_BVALID	: out std_logic;
		-- Response ready. This signal indicates that the master
    		-- can accept a write response.
		S_AXI_BREADY	: in std_logic;
		-- Read address (issued by master, acceped by Slave)
		S_AXI_ARADDR	: in std_logic_vector(C_S_AXI_ADDR_WIDTH-1 downto 0);
		-- Protection type. This signal indicates the privilege
    		-- and security level of the transaction, and whether the
    		-- transaction is a data access or an instruction access.
		S_AXI_ARPROT	: in std_logic_vector(2 downto 0);
		-- Read address valid. This signal indicates that the channel
    		-- is signaling valid read address and control information.
		S_AXI_ARVALID	: in std_logic;
		-- Read address ready. This signal indicates that the slave is
    		-- ready to accept an address and associated control signals.
		S_AXI_ARREADY	: out std_logic;
		-- Read data (issued by slave)
		S_AXI_RDATA	: out std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
		-- Read response. This signal indicates the status of the
    		-- read transfer.
		S_AXI_RRESP	: out std_logic_vector(1 downto 0);
		-- Read valid. This signal indicates that the channel is
    		-- signaling the required read data.
		S_AXI_RVALID	: out std_logic;
		-- Read ready. This signal indicates that the master can
    		-- accept the read data and response information.
		S_AXI_RREADY	: in std_logic
	);
end cc1000_v3_0_S00_AXI;

architecture arch_imp of cc1000_v3_0_S00_AXI is
  component cc1000_rxtx is
    generic(
      DBG_LEN        :    integer :=128;
      DATA_BYTES_MAX :    integer                   := 64;  -- liczba bajtow bufora ; 4*n, gdzie n=1,2,...
      PREAMBLE_BITS  :    integer                   := 96;  -- liczba bitow preambuly do wyslania ; MUSI byc liczba parzysta
      SFD_BYTES      :    std_logic_vector(0 to 15) := X"E39B"  -- 2 bajty specjalne wskazujace poczatek danych
      );
    port(
      -- wejscia do modulu od strony ukladu
      dbg_o            : out std_logic_vector(DBG_LEN-1 downto 0);
      clk_i            : in std_logic;      -- zegar z magistrali OPB
      reset_i          : in std_logic;      -- reset synchroniczny z magistrali OPB
      
      transceiver_data_i     : in std_logic_vector(0 to 31);  -- dane do wyslania (MSB na pozycji 0)
      tx_mode_i : in std_logic;          -- RX(1): DIO jest wejsciem (tryb uzywany tez, gdy jest PowerDown CC1000)
                                            -- TX(0): DIO jest wyjsciem i wysylany jest data_i
      
      start_next_rx_i : in std_logic;       -- po zakonczonym odbiorze wznawia kolejny - impuls powinien byc jak najkrotszy
      start_next_tx_i : in std_logic;       -- po zakonczonej transmisji wznawia kolejna - impuls powinien byc jak najkrotszy
      
      -- wejscia do i wyjscia z modulu od strony swiata
      dclk_i : in    std_logic;             -- ** zegar taktujacy dane ; generowany przez CC1000
      dio_io  : inout std_logic;            -- ** linia danych do/z CC1000
      
      -- wyjscia z modulu od strony ukladu
      adc_request_o : out std_logic;        -- zadanie wykonania przetwarzania przez przetwornik
      
      transceiver_data_o : out std_logic_vector(0 to 31);  -- odebrane dane (MSB na pozycji 0)
      
      rx_done_o : out std_logic;            -- wskazuje, ze zakonczono odbior
      tx_done_o : out std_logic;            -- wskazuje, ze zakonczono wysylanie
      
      tx_fifo_wr_i    : in  std_logic;
      tx_fifo_full_o  : out std_logic;
      tx_fifo_empty_o : out std_logic;
      rx_fifo_rd_i    : in  std_logic;
      rx_fifo_full_o  : out std_logic;
      rx_fifo_empty_o : out std_logic;
                  
      transceiver_busy_o : out std_logic                -- zajetosc przy wysylaniu/odbieraniu (gdy juz otrzymano SFD)
      );
  end component;

  component user_component_cc1000_config is
    generic (
      CLK_DIV_VAL :  integer := 3 -- dzielnik zegara glownego dla pclk; PCLK = CLK/(2*CLK_DIV_VAL)
      );
    port(
      -- wejscia do modulu od strony ukladu
      clk_i  : in std_logic; -- zegar magistrali
      reset_i : in std_logic;  -- reset asynchroniczny
      
      start_i : in std_logic;  -- sygnalizuje rozpoczecie zappisu/odczytu
      wr_i  : in std_logic; -- czy ZAPIS czy ODCZYT
      data_i : in std_logic_vector(7 downto 0); -- dane do wys?ania
      addr_i : in std_logic_vector(6 downto 0); -- adres rejestru
      
      -- wejscia do i wyjscia z modusu od strony swiata
      pdata_io : inout std_logic; -- ** linia danych
      pale_o : out std_logic; -- ** linia kontrolujaca
      pclk_o : out std_logic; -- ** zegar taktujacy przesyl danych do/z przetwornika; fmax=10MHz
      
      -- wyjscia z modulu od strony ukladu
      data_o : out std_logic_vector(7 downto 0); -- rejestr odbiorczy - aktualizowany po odebraniu calej danej
      addr_o : out std_logic_vector(6 downto 0); -- adres rejestru z ktorego czytamy - aktualizowany po odebraniu calej danej
      busy_o : out std_logic;      -- sygnalizuje zajetosc
      
      dbg_pdata_o : out std_logic

      );
  end component; 
  component user_component_mcp3201 is
    generic(
      CLK_DIV_VAL : integer := 64 -- dzielnik zegara glownego; CLK_OUT = CLK/CLK_DIV_VAL; minimum 2
      );
    port(
      -- wejscia do modulu od strony ukladu
      clk_i     : in std_logic;      -- zegar magistrali
      reset_i     : in std_logic;      -- reset asynchroniczny
      
      request_from_rx_i  : in std_logic;      -- zadanie przetworzenia od odbiornika
      request_from_prog_i  : in std_logic;    -- zadanie przetworzenia od programu
      
      -- wejscia do i wyjscia z modulu od strony swiata
      MCP3201_din_i     : in std_logic;      -- ** szeregowe wyjscie danych przetwornika
      MCP3201_clk_o     : out std_logic;     -- ** zegar taktujacy komunikacje z przetwornikiem; fmax=800kHz
      MCP3201_cs_o     : out std_logic;      -- ** aktywacja przetwornika (active low)
      
      -- wyjscia z modulu od strony ukladu
      adc_busy_o     : out std_logic;      -- przetwornik jest zajety przetwarzaniem
      adc_dout_o     : out std_logic_vector(11 downto 0); -- rejestr wyjsciowy z dana do odczytu
      
      adc_state_dbg_o: out std_logic_vector(3 downto 0)

      );
  end component;

	-- AXI4LITE signals
	signal axi_awaddr	: std_logic_vector(C_S_AXI_ADDR_WIDTH-1 downto 0);
	signal axi_awready	: std_logic;
	signal axi_wready	: std_logic;
	signal axi_bresp	: std_logic_vector(1 downto 0);
	signal axi_bvalid	: std_logic;
	signal axi_araddr	: std_logic_vector(C_S_AXI_ADDR_WIDTH-1 downto 0);
	signal axi_arready	: std_logic;
	signal axi_rdata	: std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal axi_rresp	: std_logic_vector(1 downto 0);
	signal axi_rvalid	: std_logic;

	-- Example-specific design signals
	-- local parameter for addressing 32 bit / 64 bit C_S_AXI_DATA_WIDTH
	-- ADDR_LSB is used for addressing 32/64 bit registers/memories
	-- ADDR_LSB = 2 for 32 bits (n downto 2)
	-- ADDR_LSB = 3 for 64 bits (n downto 3)
	constant ADDR_LSB  : integer := (C_S_AXI_DATA_WIDTH/32)+ 1;
	constant OPT_MEM_ADDR_BITS : integer := 1;
	------------------------------------------------
	---- Signals for user logic register space example
	--------------------------------------------------
	---- Number of Slave Registers 4
	signal slv_reg0	:std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal slv_reg1	:std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal slv_reg2	:std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal slv_reg3	:std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal slv_reg_rden	: std_logic;
	signal slv_reg_wren	: std_logic;
	signal reg_data_out	:std_logic_vector(C_S_AXI_DATA_WIDTH-1 downto 0);
	signal byte_index	: integer;
	signal aw_en	: std_logic;
	
  signal data_to_send     : std_logic_vector(31 downto 0) := (others => '0');
  signal transceiver_data_to_read  : std_logic_vector(31 downto 0);
  signal tx_mode       : std_logic                 := '1';  -- 0-RX, 1-TX  
  signal start_next_rx    : std_logic                 := '0';
  signal start_next_tx    : std_logic                 := '0';
  signal rx_done          : std_logic;
  signal tx_done          : std_logic;
  signal transceiver_busy : std_logic;

  signal tx_fifo_wr    : std_logic;
  signal tx_fifo_full  : std_logic;
  signal tx_fifo_empty : std_logic;
  signal rx_fifo_rd    : std_logic;
  signal rx_fifo_full  : std_logic;
  signal rx_fifo_empty : std_logic;
  
  signal adc_request              : std_logic; 
  signal request_from_prog        : std_logic;
  signal adc_busy                 : std_logic;
  signal adc_data                 : std_logic_vector(11 downto 0);
  signal data_never_read          : std_logic;   
  signal adc_busy_prev            : std_logic;
  signal slv_reg_read_select_prev : std_logic;  

  signal config_data              : std_logic_vector(7 downto 0);
  signal config_addr              : std_logic_vector(6 downto 0);
  signal config_busy              : std_logic;  
  signal config_start             : std_logic;
  signal S_AXI_ARESETN_n          : std_logic;
  
  signal config_dbg_pdata         : std_logic;
  signal MCP3201_clk              : std_logic;
  signal MCP3201_cs               : std_logic;
  signal MCP_state_dbg: std_logic_vector(3 downto 0);  
  constant  HW_VER: std_logic_vector(3 downto 0) := "0100";	

begin
  CC1000_RXTX_1  : cc1000_rxtx
    generic map (
      DBG_LEN => DBG_LEN,      
      DATA_BYTES_MAX => DATA_BYTES_MAX,
      PREAMBLE_BITS  => PREAMBLE_BITS,
      SFD_BYTES      => SFD_BYTES
      )
    port map (
      dbg_o  => dbg_o,       
      
      clk_i               => S_AXI_ACLK,
      reset_i             => S_AXI_ARESETN_n,

      transceiver_data_i  => data_to_send,
      tx_mode_i           => tx_mode,

      start_next_rx_i     => start_next_rx,
      start_next_tx_i     => start_next_tx,

      dclk_i              => CC1000_DCLK_i,
      dio_io              => CC1000_DIO_io,
      
      adc_request_o       => ADC_Request,

      transceiver_data_o  => transceiver_data_to_read,

      rx_done_o           => rx_done,
      tx_done_o           => tx_done,

      tx_fifo_wr_i        => tx_fifo_wr,
      tx_fifo_full_o      => tx_fifo_full,
      tx_fifo_empty_o     => tx_fifo_empty,
      rx_fifo_rd_i        => rx_fifo_rd,
      rx_fifo_full_o      => rx_fifo_full,
      rx_fifo_empty_o     => rx_fifo_empty,
      transceiver_busy_o  => transceiver_busy
      );    
  
  CC1000_CONFIG_1 : user_component_cc1000_config
    generic map (
      CLK_DIV_VAL => CLK_DIV_CC1000
      )
    port map (
      clk_i         => S_AXI_ACLK,
      reset_i       => S_AXI_ARESETN_n,
      -----------------------------------------
      -- |0     15|    16   |17    23|24    31|
      -----------------------------------------
      -- | unused |wr/status| addr_i | data_i |
      -----------------------------------------

      start_i  => config_start,
      wr_i     => slv_reg2(15),
      addr_i   => slv_reg2(14 downto 8),
      data_i   => slv_reg2(7 downto 0), 

      pdata_io => CC1000_PDATA_io,
      pale_o   => CC1000_PALE_o,
      pclk_o   => CC1000_PCLK_o,

      data_o   => config_data,
      addr_o   => config_addr,
      busy_o   => config_busy,
      
      dbg_pdata_o => config_dbg_pdata

      ); 
  
  
  ADC_MCP3201 : user_component_mcp3201
    generic map (
      CLK_DIV_VAL       => CLK_DIV_ADC
      )
    port map (
      clk_i               => S_AXI_ACLK,
      reset_i             => S_AXI_ARESETN_n,
      request_from_rx_i   => adc_request,
      request_from_prog_i => request_from_prog,
      MCP3201_din_i       => MCP3201_data_i,
      MCP3201_clk_o       => MCP3201_clk,
      MCP3201_cs_o        => MCP3201_cs,
      adc_busy_o              => adc_busy,
      adc_dout_o              => adc_data,
      
      adc_state_dbg_o         => MCP_state_dbg
      );

	-- I/O Connections assignments

	S_AXI_AWREADY	<= axi_awready;
	S_AXI_WREADY	<= axi_wready;
	S_AXI_BRESP	<= axi_bresp;
	S_AXI_BVALID	<= axi_bvalid;
	S_AXI_ARREADY	<= axi_arready;
	S_AXI_RDATA	<= axi_rdata;
	S_AXI_RRESP	<= axi_rresp;
	S_AXI_RVALID	<= axi_rvalid;
	-- Implement axi_awready generation
	-- axi_awready is asserted for one S_AXI_ACLK clock cycle when both
	-- S_AXI_AWVALID and S_AXI_WVALID are asserted. axi_awready is
	-- de-asserted when reset is low.

	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      axi_awready <= '0';
	      aw_en <= '1';
	    else
	      if (axi_awready = '0' and S_AXI_AWVALID = '1' and S_AXI_WVALID = '1' and aw_en = '1') then
	        -- slave is ready to accept write address when
	        -- there is a valid write address and write data
	        -- on the write address and data bus. This design 
	        -- expects no outstanding transactions. 
	           axi_awready <= '1';
	           aw_en <= '0';
	        elsif (S_AXI_BREADY = '1' and axi_bvalid = '1') then
	           aw_en <= '1';
	           axi_awready <= '0';
	      else
	        axi_awready <= '0';
	      end if;
	    end if;
	  end if;
	end process;

	-- Implement axi_awaddr latching
	-- This process is used to latch the address when both 
	-- S_AXI_AWVALID and S_AXI_WVALID are valid. 

	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      axi_awaddr <= (others => '0');
	    else
	      if (axi_awready = '0' and S_AXI_AWVALID = '1' and S_AXI_WVALID = '1' and aw_en = '1') then
	        -- Write Address latching
	        axi_awaddr <= S_AXI_AWADDR;
	      end if;
	    end if;
	  end if;                   
	end process; 

	-- Implement axi_wready generation
	-- axi_wready is asserted for one S_AXI_ACLK clock cycle when both
	-- S_AXI_AWVALID and S_AXI_WVALID are asserted. axi_wready is 
	-- de-asserted when reset is low. 

	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      axi_wready <= '0';
	    else
	      if (axi_wready = '0' and S_AXI_WVALID = '1' and S_AXI_AWVALID = '1' and aw_en = '1') then
	          -- slave is ready to accept write data when 
	          -- there is a valid write address and write data
	          -- on the write address and data bus. This design 
	          -- expects no outstanding transactions.           
	          axi_wready <= '1';
	      else
	        axi_wready <= '0';
	      end if;
	    end if;
	  end if;
	end process; 
	
	-- Implement write response logic generation
	-- The write response and response valid signals are asserted by the slave 
	-- when axi_wready, S_AXI_WVALID, axi_wready and S_AXI_WVALID are asserted.  
	-- This marks the acceptance of address and indicates the status of 
	-- write transaction.

	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      axi_bvalid  <= '0';
	      axi_bresp   <= "00"; --need to work more on the responses
	    else
	      if (axi_awready = '1' and S_AXI_AWVALID = '1' and axi_wready = '1' and S_AXI_WVALID = '1' and axi_bvalid = '0'  ) then
	        axi_bvalid <= '1';
	        axi_bresp  <= "00"; 
	      elsif (S_AXI_BREADY = '1' and axi_bvalid = '1') then   --check if bready is asserted while bvalid is high)
	        axi_bvalid <= '0';                                 -- (there is a possibility that bready is always asserted high)
	      end if;
	    end if;
	  end if;                   
	end process; 

	-- Implement axi_arready generation
	-- axi_arready is asserted for one S_AXI_ACLK clock cycle when
	-- S_AXI_ARVALID is asserted. axi_awready is 
	-- de-asserted when reset (active low) is asserted. 
	-- The read address is also latched when S_AXI_ARVALID is 
	-- asserted. axi_araddr is reset to zero on reset assertion.

	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      axi_arready <= '0';
	      axi_araddr  <= (others => '1');
	    else
	      if (axi_arready = '0' and S_AXI_ARVALID = '1') then
	        -- indicates that the slave has acceped the valid read address
	        axi_arready <= '1';
	        -- Read Address latching 
	        axi_araddr  <= S_AXI_ARADDR;           
	      else
	        axi_arready <= '0';
	      end if;
	    end if;
	  end if;                   
	end process; 

	-- Implement axi_arvalid generation
	-- axi_rvalid is asserted for one S_AXI_ACLK clock cycle when both 
	-- S_AXI_ARVALID and axi_arready are asserted. The slave registers 
	-- data are available on the axi_rdata bus at this instance. The 
	-- assertion of axi_rvalid marks the validity of read data on the 
	-- bus and axi_rresp indicates the status of read transaction.axi_rvalid 
	-- is deasserted on reset (active low). axi_rresp and axi_rdata are 
	-- cleared to zero on reset (active low).  
	process (S_AXI_ACLK)
	begin
	  if rising_edge(S_AXI_ACLK) then
	    if S_AXI_ARESETN = '0' then
	      axi_rvalid <= '0';
	      axi_rresp  <= "00";
	    else
	      if (axi_arready = '1' and S_AXI_ARVALID = '1' and axi_rvalid = '0') then
	        -- Valid read data is available at the read data bus
	        axi_rvalid <= '1';
	        axi_rresp  <= "00"; -- 'OKAY' response
	      elsif (axi_rvalid = '1' and S_AXI_RREADY = '1') then
	        -- Read data is accepted by the master
	        axi_rvalid <= '0';
	      end if;            
	    end if;
	  end if;
	end process;
	
		-- Implement memory mapped register select and write logic generation
	-- The write data is accepted and written to memory mapped registers when
	-- axi_awready, S_AXI_WVALID, axi_wready and S_AXI_WVALID are asserted. Write strobes are used to
	-- select byte enables of slave registers while writing.
	-- These registers are cleared when reset (active low) is applied.
	-- Slave register write enable is asserted when valid address and data are available
	-- and the slave is ready to accept the write address and write data.
	slv_reg_wren <= axi_wready and S_AXI_WVALID and axi_awready and S_AXI_AWVALID ;

	process (S_AXI_ACLK)
	variable loc_addr :std_logic_vector(OPT_MEM_ADDR_BITS downto 0); 
	begin
	  if rising_edge(S_AXI_ACLK) then 
	    if S_AXI_ARESETN = '0' then
	      slv_reg0 <= (others => '0');
	      slv_reg1 <= (others => '0');
	      slv_reg2 <= (others => '0');
	      slv_reg3 <= (others => '0');
          data_to_send      <= (others => '0');
          tx_fifo_wr        <= '0';
          tx_mode           <= '0';
          start_next_rx     <= '0';
          start_next_tx     <= '0';
          rx_fifo_rd        <= '0';      
          config_start      <= '0';
          request_from_prog <= '0';          
	    else
          tx_fifo_wr        <= '0';
          start_next_rx     <= '0';
          start_next_tx     <= '0';
          rx_fifo_rd        <= '0';
          config_start      <= '0';
          request_from_prog <= '0';
	      loc_addr := axi_awaddr(ADDR_LSB + OPT_MEM_ADDR_BITS downto ADDR_LSB);	      
	      if (slv_reg_wren = '1') then	        	        
	        case loc_addr is
	          -- write to DATA_REG:
	          when b"00" =>
	            for byte_index in 0 to (C_S_AXI_DATA_WIDTH/8-1) loop
	              if ( S_AXI_WSTRB(byte_index) = '1' ) then
	                -- Respective byte enables are asserted as per write strobes                   
	                -- slave register 0
                    data_to_send(byte_index*8+7 downto byte_index*8) <= S_AXI_WDATA(byte_index*8+7 downto byte_index*8);
	              end if;                  
	            end loop;
	            tx_fifo_wr <= '1';
		      -- write to STATUS_CONTROL_REG:
	          when b"01" =>
	            for byte_index in 0 to (C_S_AXI_DATA_WIDTH/8-1) loop
	              if ( S_AXI_WSTRB(byte_index) = '1' ) then
	                -- Respective byte enables are asserted as per write strobes                   
	                -- slave register 1
	                slv_reg1(byte_index*8+7 downto byte_index*8) <= S_AXI_WDATA(byte_index*8+7 downto byte_index*8);
	              end if;
	            end loop;
                if (S_AXI_WSTRB(0) = '1') then
                  if S_AXI_WDATA(1) = '1'  and S_AXI_WDATA(0) = '0'then
                    tx_mode    <= '0';
                  end if;
                  if S_AXI_WDATA(1) = '0'  and S_AXI_WDATA(0) = '1'then
                    tx_mode    <= '1';
                  end if;				              
                  if S_AXI_WDATA(2) = '1' then  -- gdy ustawiona maska modyfikacji
                    start_next_rx <= '1';
                  end if;                
                  if S_AXI_WDATA(3) = '1' then  -- gdy ustawiona maska modyfikacji
                    start_next_tx <= '1';
                  end if;                                
                  if S_AXI_WDATA(4) = '1' then  -- gdy ustawiona maska modyfikacji
                    rx_fifo_rd    <= '1';
                  end if;
                end if;
              -- write to CONFIG_REG:  	            
	          when b"10" =>
	            for byte_index in 0 to (C_S_AXI_DATA_WIDTH/8-1) loop
	              if ( S_AXI_WSTRB(byte_index) = '1' ) then
	                -- Respective byte enables are asserted as per write strobes                   
	                -- slave registor 2
	                slv_reg2(byte_index*8+7 downto byte_index*8) <= S_AXI_WDATA(byte_index*8+7 downto byte_index*8);
	              end if;
	            end loop;
                if ( S_AXI_WSTRB(1) = '1' ) then                                              
                  if config_busy = '0' then  -- not adc_start_conversion
                    config_start <= '1';
                  end if;
                end if;           
			  -- write to ADC_REG:	            
	          when b"11" =>
	            for byte_index in 0 to (C_S_AXI_DATA_WIDTH/8-1) loop
	              if ( S_AXI_WSTRB(byte_index) = '1' ) then
	                -- Respective byte enables are asserted as per write strobes                   
	                -- slave registor 3
	                slv_reg3(byte_index*8+7 downto byte_index*8) <= S_AXI_WDATA(byte_index*8+7 downto byte_index*8);
	              end if;
	            end loop;
			    if ( S_AXI_WSTRB(0) = '1' ) then
				  if S_AXI_WDATA(0) = '1' then
			        request_from_prog <= '1';
                  end if;
			    end if;
	          when others =>
	            slv_reg0 <= slv_reg0;
	            slv_reg1 <= slv_reg1;
	            slv_reg2 <= slv_reg2;
	            slv_reg3 <= slv_reg3;
	        end case;
	      end if;
	    end if;
	  end if;                   
	end process; 
	

	-- Implement memory mapped register select and read logic generation
	-- Slave register read enable is asserted when valid address is available
	-- and the slave is ready to accept the read address.
	slv_reg_rden <= axi_arready and S_AXI_ARVALID and (not axi_rvalid) ;

	process (transceiver_data_to_read, slv_reg0, slv_reg1, slv_reg2, slv_reg3, axi_araddr, S_AXI_ARESETN, slv_reg_rden, 
           rx_fifo_empty, rx_fifo_full, tx_fifo_empty, tx_fifo_full, tx_done, rx_done, tx_mode, transceiver_busy,
           adc_busy, data_never_read, slv_reg_read_select_prev, adc_data, config_addr, config_data, 
           config_start, request_from_prog, config_busy)
	variable loc_addr :std_logic_vector(OPT_MEM_ADDR_BITS downto 0);
	begin
	    -- Address decoding for reading registers
	    loc_addr := axi_araddr(ADDR_LSB + OPT_MEM_ADDR_BITS downto ADDR_LSB);
	    case loc_addr is
	      when b"00" =>
            reg_data_out               <= transceiver_data_to_read;
	      when b"01" =>
	        reg_data_out(31 downto 11) <=  HW_VER & "00000000000000000";
            reg_data_out(10)           <= rx_fifo_empty;
            reg_data_out(9)            <= rx_fifo_full;
            reg_data_out(8)            <= tx_fifo_empty;
            reg_data_out(7)            <= tx_fifo_full;
            reg_data_out(6)            <= tx_done;
            reg_data_out(5)            <= rx_done;
            reg_data_out(4 downto 2)   <= (others => '0');
            reg_data_out(1)            <= tx_mode;
            reg_data_out(0)            <= transceiver_busy;
	      when b"10" =>
	        reg_data_out(31 downto 16) <= (others => '0');
            reg_data_out(15)           <= config_busy;
            reg_data_out(14 downto 8)  <= config_addr;
            reg_data_out(7 downto 0)   <= config_data;
	      when b"11" =>
	        reg_data_out(31 downto 28) <= (others => '0');       
            reg_data_out(15)           <= adc_busy;
            reg_data_out(14)           <= data_never_read;         	
 	        reg_data_out(13 downto 12) <= (others => '0');       
            reg_data_out(11 downto 0)  <= adc_data;
	      when others =>
	        reg_data_out  <= (others => '0');
	    end case;
	end process; 

	-- Output register or memory read data
	process( S_AXI_ACLK ) is
	begin
	  if (rising_edge (S_AXI_ACLK)) then
	    if ( S_AXI_ARESETN = '0' ) then
	      axi_rdata  <= (others => '0');
	    else
	      if (slv_reg_rden = '1') then
	        -- When there is a valid read address (S_AXI_ARVALID) with 
	        -- acceptance of read address by the slave (axi_arready), 
	        -- output the read dada 
	        -- Read address mux
	          axi_rdata <= reg_data_out;     -- register read data
	      end if;   
	    end if;
	  end if;
	end process;


	-- Add user logic here
  DATA_NEVER_READ_PROC : process( S_AXI_ACLK ) is
    variable loc_addr_v :std_logic_vector(OPT_MEM_ADDR_BITS downto 0);
  begin
    if S_AXI_ACLK'event and S_AXI_ACLK = '1' then
      if S_AXI_ARESETN = '1' then
        adc_busy_prev            <= '0';
        slv_reg_read_select_prev <= '0';
        data_never_read          <= '0';
      else
        loc_addr_v := axi_araddr(ADDR_LSB + OPT_MEM_ADDR_BITS downto ADDR_LSB);
        if adc_busy_prev = '1' and adc_busy = '0' then
          data_never_read        <= '1';
        elsif slv_reg_read_select_prev = '1' and slv_reg_rden = '1' and loc_addr_v="10" then
          data_never_read        <= '0';
        else
          data_never_read        <= data_never_read;
        end if;
        adc_busy_prev            <= adc_busy;
        if (slv_reg_rden = '1' and loc_addr_v="10") then  
          slv_reg_read_select_prev <= '1';
        else
          slv_reg_read_select_prev <= '0';
        end if;
      end if;
    end if;
  end process DATA_NEVER_READ_PROC;

  
  S_AXI_ARESETN_n <= not S_AXI_ARESETN;   
  MCP3201_clk_o <= MCP3201_clk;
  MCP3201_cs_o <= MCP3201_cs;   
	-- User logic ends

end arch_imp;
