-- Modul nadajnika i odbiornika danych wysylanych do lub odbieranych z CC1000. Napisany w ramach pracy magisterskiej na WETI PG.
--
-- Autor: Bartosz Marcin Durski, korekta i adaptacja do Zybo: Marek Wojcikowski 
--
-- !!! WAzNE !!! Aby nie spalic linii dwukierunkowej DIO.
-- Przelaczac 'rx_tx_mode' na '0' (czyli RX) przed przelaczeniem CC1000 w tryb RX.
-- Przelaczac 'rx_tx_mode' na '1' (czyli TX) dopiero po przelaczeniu CC1000 w tryb TX.
--
-- zadanie do ADC:
-- Po wykryciu SFD wysylane jest zadanie przetwarzania do przetwornika ADC (ktore trwa 1 cykl 'clk').
--
-- Tryb TX:
-- Po przejsciu w tryb TX, uklad natychmiast uruchamia wysyl. Wysyla preambule, SFD, i tyle bajtow ile jest
-- podanych w 1. bajcie danych lub tyle ile miesciu bufor, jezeli w 1. bajcie jest podana wartosc wieksza od wielkosci
-- buforu. Po wyslaniu wszystkich danych ustawiany jest 'tx_done' i trwa do czasu 'start_next_tx', 'resetu' lub
-- przejscia w tryb RX. 'tx_done' moze byc uzyty do generowania przerwania zakonczenia wysylania, ale nalezy
-- pamietac, ze 'tx_done' trwa tak dlugo jak wczesniej przedstawiono. Dlatego nalezy generowac przerwanie na
-- pojawienie sie narastajacego zbocza, bo przerwanie na poziom, by powodowalo kolejne wchodzenie do funcji obslugi
-- przerwania.
--
-- Tryb RX:
-- Sprawdza 4 kolejne odebrane bajty. Jezeli 4 kolejne bajty beda rowne kolejno dwom bajtom preambuly i 2 bajtom SFD,
-- to przechodzi do odbioru dannych. Po odbiorze 1. bajtu danych sprawdzamy ile bajtow danych jest przesylanych
-- i jezeli jest wiecej niz DATA_BYTES_MAX, to powracamy na poczatek algorytmu. Po odebraniu wszystkich danych
-- ustawiany jest 'rx_done' i trwa do czasu 'start_next_rx', 'resetu' lub przejscia w tryb TX. Po odebraniu calej
-- ramki blokowane jest odbieranie nastepnych ramek, aby nie nadpisac bufora (data_o). Odblokowac mozna za
-- pomoca 'reseta', 'start_next_rx' lub przejscia w tryb TX. 'rx_done' moze byc uzyty do generowania przerwania
-- zakonczenia wysylania, ale nalezy pamietac, ze 'rx_done' trwa tak dlugo jak wczesniej przedstawiono.
-- Dlatego nalezy generowac przerwanie na pojawienie sie narastajacego zbocza, bo przerwanie na poziom, by
-- powodowalo kolejne wchodzenie do funcji obslugi przerwania.
--
-- Busy:
-- Gdy jest tryb TX, to busy jest ustawione od chwili rozpoczecia wysylania do chwili wyslania ostatniego bitu.
-- Gdy jest tryb RX, to busy jest ustawione od momentu odebrania SFD do momentu odebrania calej ramki.
--
-- start_next_tx:
-- Po zakonczonej transmisji wysylane jest caly czas zero, aby wznowic transmisje mozna m.in. skorzystac ze
-- 'start_next_tx'. Ustawienie na '1' przygotowuje do ponownej transmisji, a automatyczne zejscie na '0'
-- w nastepnym cyklu zegara uruchamia ja. Nie powinno sie korzystac ze 'start_next_tx', gdy trwa jeszcze wysylanie.
-- Co nie znaczy, ze nie mozna.
--
-- start_next_rx:
-- Po zakonczonym odbiorze blokowany jest odbior kolejnych ramek. Aby odblokowac mozna m.in. skorzystac ze
-- 'start_next_rx'. Ustawienie na '1' przygotowuje do ponownego odbioru, a automatyczne zejscie na '0'
-- w nastepnym cyklu zegara uruchamia odbior. Nie powinno sie korzystac ze 'start_next_rx', gdy nie jest
-- wlaczone blokowanie odbioru kolejnych ramek. Co nie znaczy, ze nie mozna.
--
-- ** = sygnal wychodzi z lub wchodzi do ukladu

library ieee;
use ieee.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

library UNISIM;
use UNISIM.VCOMPONENTS.all;

entity cc1000_rxtx is
    generic(
        DBG_LEN        :    integer :=128;
        DATA_BYTES_MAX :    integer                   := 64;  -- liczba bajtow bufora ; 4*n, gdzie n=1,2,...
        PREAMBLE_BITS  :    integer                   := 96;  -- liczba bitow preambuly do wyslania ; MUSI byc liczba parzysta
        SFD_BYTES      :    std_logic_vector(0 to 15) := X"E39B"  -- 2 bajty specjalne wskazujace poczatek danych     
    );
    port(
        -- wejscia do modulu od strony ukladu
        dbg_o            : out std_logic_vector(DBG_LEN-1 downto 0);
        clk_i            : in std_logic;      -- zegar z magistrali OPB
        reset_i          : in std_logic;      -- reset synchroniczny z magistrali OPB
    
        transceiver_data_i     : in std_logic_vector(0 to 31);  -- dane do wyslania (MSB na pozycji 0)
        tx_mode_i : in std_logic;       -- RX=0: DIO jest wejsciem (tryb uzywany tez, gdy jest PowerDown CC1000)
                                        -- TX=1: DIO jest wyjsciem i wysylany jest data_i

        start_next_rx_i : in std_logic;  -- po zakonczonym odbiorze wznawia kolejny 
        start_next_tx_i : in std_logic;  -- po zakonczonej transmisji wznawia kolejna

        -- wejscia do i wyjscia z modulu od strony swiata
        dclk_i : in    std_logic;             -- ** zegar taktujacy dane ; generowany przez CC1000
        dio_io  : inout std_logic;             -- ** linia danych do/z CC1000

        -- wyjscia z modulu od strony ukladu
        adc_request_o : out std_logic;        -- zadanie wykonania przetwarzania przez przetwornik
    
        transceiver_data_o : out std_logic_vector(0 to 31);  -- odebrane dane (MSB na pozycji 0)
        tx_done_o : out std_logic;            -- wskazuje, ze zakonczono wysylanie
        rx_done_o : out std_logic;            -- wskazuje, ze zakonczono odbior
        
        tx_fifo_wr_i    : in  std_logic;
        tx_fifo_full_o  : out std_logic;
        tx_fifo_empty_o : out std_logic;    
               
        rx_fifo_rd_i    : in  std_logic;
        rx_fifo_full_o  : out std_logic;
        rx_fifo_empty_o : out std_logic;
        
        transceiver_busy_o : out std_logic                -- zajetosc przy wysylaniu/odbieraniu (gdy juz otrzymano SFD)
    );
end entity;


architecture cc1000_rxtx_arch of cc1000_rxtx is  
    COMPONENT fifo_32_512_sync
      PORT (
        clk : IN STD_LOGIC;
        srst : IN STD_LOGIC;
        din : IN STD_LOGIC_VECTOR(31 DOWNTO 0);
        wr_en : IN STD_LOGIC;
        rd_en : IN STD_LOGIC;
        dout : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
        full : OUT STD_LOGIC;
        empty : OUT STD_LOGIC
      );
    END COMPONENT;    

    function max3(a : integer; b : integer; c : integer)
        return integer;

    function max3(a : integer; b : integer; c : integer)
        return integer is
    begin
        if a > b then
            if a > c then
                return a;
            else
                return c;
            end if;
        else
            if b > c then
                return b;
            else
                return c;
            end if;
        end if;
    end max3;

    -- Bufor 3-stanowy
    component IOBUF is
    port (
      I  : in    std_logic;
      IO : inout std_logic;
      O  : out   std_logic;
      T  : in    std_logic
      );
    end component;

    signal data_o_tmp : std_logic_vector(0 to 31);  -- ze wzgledu na to, ze data_o jest wyjsciem, a (moze) bedziemy jednak czytac
    
    signal tri_out   : std_logic := '0';  -- sygnal wychodzacy przez bufor 3-stanowy
    signal tri_in    : std_logic;         -- sygnal przychodzacy przez bufor 3-stanowy
    signal tri_state : std_logic;         -- sygnal sterujacy kierunkiem buforu (0-wyjscie, 1-wejscie)
    
    signal rx_busy : std_logic;           -- wskazuje, ze odbior trwa (pojawilo sie juz wczesniej SFD)
    signal tx_busy : std_logic;           -- wskazuje, ze transmisja trwa
    
    signal adc_request_now  : std_logic;  -- gdy pojawi sie SFD, to jest ustawiane i zerowane w nastepnym cyklu zegara
    signal adc_request_prev : std_logic;  -- zwiera poprzedni stan 'adc_request_now' sprzed 1 cyklu 'clk'
    
    signal rx_done : std_logic := '0';  -- ze wzgledu na to, ze rx_done jest wyjsciem, a (moze) bedziemy jednak czytac
    signal tx_done : std_logic := '0';  -- ze wzgledu na to, ze tx_done jest wyjsciem, a (moze) bedziemy jednak czytac
    
    signal sfd_buf : std_logic_vector(0 to 8*4-1) := (others => '0');  -- sluzy do sprawdzania czy odebrano 2 bajty preambuly i SFD ; bufor FIFO

    -- PREAMBLE - sprawdzanie czy pojawily sie kolejno 2 bajty preambuly (0xAA 0xAA) i 2 bajty specjalne (SFD_BYTES)
    -- DATA - odbieranie danych
    -- WAITING - blokowanie odbioru nowych ramek i czekanie na 'reset', 'start_next_rx' lub zmiane trybu na TX
    type RX_states is (PREAMBLE, DATA, WAITING);
    signal rx_state : RX_states := PREAMBLE;  -- opisuje stany w jakich znajduje sie proces odbierania

    -- PREAMBLE - wysylanie preambuly
    -- SFD - wysylanie znakow specjalnych
    -- DATA - wysylanie danych
    -- WAITING - wysylanie samych zer i czekanie na 'reset', 'start_next_tx' lub zmiane trybu na RX
    type TX_states is (PREAMBLE, SFD, DATA, WAITING);
    signal tx_state : TX_states := PREAMBLE;  -- opisuje stany w jakich znajduje sie proces nadawania
    
    signal tx_fifo_rst     : std_logic;
    signal tx_fifo_full    : std_logic;
    signal tx_fifo_rd      : std_logic;
    signal data_for_cc1000r : std_logic_vector(31 downto 0);
    signal tx_fifo_empty   : std_logic;
    
    signal rx_fifo_rst          : std_logic;
    signal rx_fifo_wr           : std_logic;    
    signal data_from_cc1000r     : std_logic_vector(31 downto 0);
    signal rx_fifo_full         : std_logic;
    signal rx_fifo_empty        : std_logic;

    signal dclk_prev0           : std_logic;
    signal dclk_prev1           : std_logic;
    signal dclk_sync            : std_logic;
    
    signal start_next_tx_sync   : std_logic;
    
    signal start_next_rx_prev   : std_logic;
    signal start_next_rx_led    : std_logic;
    signal dclk_led             : std_logic;  --leading edge detector
    signal dclk_fed             : std_logic;  --falling edge detector     
begin
    ----------------------------------------------------------------------------------  
    dbg_o(15 downto 0)   <= sfd_buf(16 to 8*4-1);
    --dbg_o(23 downto 16) <= conv_std_logic_vector(tx_counter_bytes_v, 8);
    --dbg_o(31 downto 24) <= tx_data_len_v;    
    dbg_o(63 downto 32)  <= data_from_cc1000r;    
    dbg_o(64) <= tri_in;
    dbg_o(65) <= rx_fifo_rst;
    dbg_o(66) <= rx_fifo_wr;
    dbg_o(67) <= tx_mode_i;               
    dbg_o(68) <= start_next_rx_i;
    dbg_o(69) <= rx_busy;
    dbg_o(70) <= rx_done;
    dbg_o(71) <= dclk_i;   
    dbg_o(72) <= rx_fifo_full;
    dbg_o(73) <= rx_fifo_empty;
    dbg_o(74) <= rx_fifo_rd_i; 
    dbg_o(75) <= '0'; --
    --dbg_o(81 downto 76) <= conv_std_logic_vector(rx_counter_bits_v, 6);
    --dbg_o(87 downto 82) <= conv_std_logic_vector(rx_counter_bytes_v, 6);
    --dbg_o(93 downto 88) <= conv_std_logic_vector(rx_data_len_v, 6);
    process (rx_state) 
    begin
        case (rx_state) is
            when PREAMBLE => dbg_o(96 downto 94) <= "001";
            when DATA     => dbg_o(96 downto 94) <= "010";
            when WAITING  => dbg_o(96 downto 94) <= "100";
            when others  =>  dbg_o(96 downto 94) <= "000";
        end case;
    end process;
  
    process (tx_state) 
    begin
        case (tx_state) is
            when PREAMBLE => dbg_o(99 downto 97) <= "001";
            when SFD      => dbg_o(99 downto 97) <= "010";
            when DATA     => dbg_o(99 downto 97) <= "100";
            when WAITING  => dbg_o(99 downto 97) <= "111";
            when others   => dbg_o(99 downto 97) <= "000";
        end case;
    end process;
    dbg_o(100)  <= start_next_tx_i;
    dbg_o(101)  <= start_next_tx_sync;
    dbg_o(102)  <= tx_fifo_wr_i;
    dbg_o(103) <= tx_fifo_rd;
    dbg_o(104) <= tri_out;
    dbg_o(105) <= tx_done;
    dbg_o(106) <= tx_fifo_full;
    dbg_o(107) <= tx_fifo_empty;  
    dbg_o(108) <= tx_fifo_rst;
    dbg_o(109) <= tx_fifo_rd;  
    --dbg_o(115 downto 110) <= conv_std_logic_vector(tx_counter_bits_v, 6);    
    dbg_o(116) <= reset_i;
    dbg_o(117) <= start_next_rx_led;
    dbg_o(118) <= dclk_led;
    dbg_o(119) <= dclk_fed;        
    dbg_o(DBG_LEN-1 downto 120) <= data_for_cc1000r(7 downto 0);                             
    ----------------------------------------------------------------------------------
    -- Bufor 3-stanowy
    iobuf_0 : IOBUF
    port map (
        I  => tri_out,
        IO => dio_io,
        O  => tri_in,
        T  => tri_state
    );
           
    TX_FIFO : fifo_32_512_sync port map (
        srst   => tx_fifo_rst,
        clk    => clk_i,
        din    => transceiver_data_i,
        wr_en  => tx_fifo_wr_i,    
        rd_en  => tx_fifo_rd,
        dout   => data_for_cc1000r,
        full   => tx_fifo_full,
        empty  => tx_fifo_empty
    );

    RX_FIFO : fifo_32_512_sync port map (
        srst   => rx_fifo_rst,
        clk    => clk_i,    
        din    => data_from_cc1000r,
        wr_en  => rx_fifo_wr,    
        rd_en  => rx_fifo_rd_i,
        dout   => data_o_tmp,
        full   => rx_fifo_full,
        empty  => rx_fifo_empty    
    );
             
    tx_fifo_full_o  <= tx_fifo_full;
    tx_fifo_empty_o <= tx_fifo_empty;
    rx_fifo_full_o  <= rx_fifo_full;
    transceiver_data_o <= data_o_tmp;  
    
    transceiver_busy_o      <= rx_busy or tx_busy;
    tx_busy   <= '1' when (tx_done = '0' and reset_i = '0' and tx_mode_i = '1' and start_next_tx_sync = '0') else
               '0';
    tri_state      <= not tx_mode_i;  
    rx_done_o      <= rx_done;
    tx_done_o      <= tx_done;
    
    dclk_led <= not dclk_sync and dclk_prev1;  
    dclk_fed <= dclk_sync and not dclk_prev1;
    DCLK_SYNC_PROC : process(clk_i, reset_i) 
    begin
        if (reset_i = '1') then        
            dclk_prev0 <= '0';
            dclk_prev1 <= '0';
            dclk_sync <= '0';
        elsif rising_edge(clk_i) then
            dclk_prev0 <= dclk_i;
            dclk_prev1 <= dclk_prev0;
            dclk_sync <= dclk_prev1;
        end if;
    end process;         

    -- Proces generujacy impuls nakazujacy przetwornikowi ADC wykonac przetwarzanie
    -- Zasada dzialania polega na sprawdzaniu stanu obencego i poprzedniego. Jezeli obecny =1 i poprzedni =0, to
    -- nalezy wyslac zadanie do przetwornika. Poprzedni stan nadpisujemy obecnym.
    ADC_REQUEST_PROC : process(clk_i)
    begin
        if rising_edge(clk_i) then
            if adc_request_now = '1' and adc_request_prev = '0' then
                adc_request_o    <= '1';
            else
                adc_request_o    <= '0';
            end if;
            adc_request_prev <= adc_request_now;
        end if;
    end process ADC_REQUEST_PROC;

    TX_START_SYNC : process(clk_i, reset_i) 
    begin
        if (reset_i = '1') then
            start_next_tx_sync <= '0';        
        elsif rising_edge(clk_i) then
            if (start_next_tx_i = '1') then 
                start_next_tx_sync <= '1';
            end if;
            if (start_next_tx_sync = '1' and dclk_sync='1' and dclk_i='0') then
                start_next_tx_sync <= '0';
            end if;          
        end if;
    end process;

    -- Proces wysylajacy dane do CC1000
    TX_PROC                   : process(clk_i, reset_i)
        variable tx_counter_bits_v : integer range 0 to max3(PREAMBLE_BITS, 8*DATA_BYTES_MAX, 16)-1 := 0;  -- odlicza bity: preambuly lub danych lub SFD ; 16, bo 16 bitow SFD
        variable tx_counter_bytes_v : integer range 0 to 255;
        variable tx_data_len_v     : integer range 0 to 255;
    begin
        if (reset_i = '1') then  -- gdy reset lub wlaczony tryb RX lub wznow nadawanie
            tx_state           <= PREAMBLE;
            tri_out            <= '0';
            tx_done            <= '0';
            tx_counter_bits_v  := 0;
            tx_counter_bytes_v := 0;
            tx_data_len_v      := 0;
            tx_fifo_rd         <= '0';
            tx_fifo_rst        <= '1';
        else    
            if rising_edge(clk_i) then
                tx_fifo_rst        <= '0';
                tx_fifo_rd         <= '0';
                if (tx_mode_i = '0') or (start_next_tx_i = '1') then
                    tx_state           <= PREAMBLE;
                    tri_out            <= '0';
                    tx_done            <= '0';
                    tx_counter_bits_v  := 0;                                   
                    tx_counter_bytes_v := 0;
                    tx_data_len_v      := 0;
                else
                    if (dclk_fed = '1') then
                        case tx_state is
                            when PREAMBLE =>              
                                tri_out      <= not tri_out;
                                if tx_counter_bits_v >= PREAMBLE_BITS-1 then  -- wyslij ostatni bit preambuly i przejdz do wysylania SFD
                                    tx_state   <= SFD;
                                    tx_counter_bits_v  := 0;
                                else
                                    tx_counter_bits_v  := tx_counter_bits_v + 1;
                                end if;              
                            when SFD =>                    
                                tri_out      <= SFD_BYTES(tx_counter_bits_v);
                                if tx_counter_bits_v >= 16-1 then
                                    tx_state   <= DATA;
                                    tx_counter_bits_v := 0;
                                    tx_counter_bytes_v := 0;
                                    tx_fifo_rd <= '1';
                                else
                                    tx_counter_bits_v := tx_counter_bits_v + 1;
                                end if;                       
                            when DATA =>
                                tri_out <= data_for_cc1000r(tx_counter_bits_v mod 32);
                                if tx_counter_bits_v = 1 then                          
                                    tx_data_len_v := conv_integer(data_for_cc1000r(7 downto 0));
                                end if;        
                                if (tx_counter_bits_v mod 32 = 31) and (tx_fifo_empty = '0') then
                                    tx_fifo_rd <= '1';
                                end if;
                                -- gdy wyslano wszystkie dane:
                                if (((tx_counter_bits_v+1) mod 8 = 0) and (tx_counter_bytes_v = tx_data_len_v)) or (tx_counter_bits_v >= 8*DATA_BYTES_MAX-1) then                                                                
                                    tx_state <= WAITING;                                    
                                else
                                    tx_counter_bits_v := tx_counter_bits_v + 1;
                                    if (tx_counter_bits_v mod 8 = 0 ) then
                                        tx_counter_bytes_v := tx_counter_bytes_v + 1;
                                    end if;
                                end if;
                            when WAITING =>
                                tri_out     <= '0';
                                tx_done     <= '1';
                                tx_fifo_rst <= '1';                
                        end case;
                    end if;
                end if;
                dbg_o(115 downto 110) <= conv_std_logic_vector(tx_counter_bits_v, 6);
                dbg_o(31 downto 24)   <= conv_std_logic_vector(tx_data_len_v, 8);
                dbg_o(23 downto 16) <= conv_std_logic_vector(tx_counter_bytes_v, 8);
            end if;
        end if;
    end process TX_PROC;

    -- Proces odbierajacy dane z CC1000
    RX_PROC                    : process(dclk_i, reset_i)
        variable rx_counter_bits_v  : integer range 0 to 32               := 0;  -- odlicza kolejne bity danych do odebrania
        variable rx_counter_bytes_v : integer range 0 to 255 := 0;  -- odlicza kolejne bajty danych do odebrania
        variable rx_data_len_v      : integer range 0 to 255;
        variable rx_fifo_write_v : boolean := FALSE;
    begin
    if (reset_i = '1') then  -- gdy reset lub wlaczony TX lub wznow odbior
        rx_state                <= PREAMBLE;
        rx_busy                 <= '0';
        rx_done                 <= '0';
        adc_request_now         <= '0';
        rx_counter_bits_v        := 0;
        rx_counter_bytes_v       := 0;
        rx_data_len_v            := 0;  
        rx_fifo_rst             <= '1';
        rx_fifo_wr              <= '0';
    elsif rising_edge(clk_i) then        
        rx_fifo_rst           <= '0';
        rx_fifo_wr            <= '0';
        rx_fifo_write_v := FALSE;
        if (tx_mode_i = '1') or (start_next_rx_i = '1') then
            rx_state                <= PREAMBLE;
            rx_busy                 <= '0';
            rx_done                 <= '0';
            adc_request_now         <= '0';
            rx_counter_bits_v        := 0;
            rx_counter_bytes_v       := 0;
            rx_data_len_v            := 0;  
            rx_fifo_rst             <= '1';
            rx_fifo_wr              <= '0';
        else
            if (dclk_led = '1') then
                case rx_state is
                    when PREAMBLE                  =>                                 
                        if (sfd_buf(1 to 31) & tri_in) = (X"AAAA" & SFD_BYTES) then  -- (1 to 31) a nie (0 to 30), bo w jeszcze rejestry bufory nie sa przesuniete
                            rx_state        <= DATA;
                            rx_busy         <= '1';
                            adc_request_now <= '1';   -- ustaw zadanie dla przetwornik
                            sfd_buf         <= (others => '0');                      
                        else
                            for idx in 0 to 30 loop
                                sfd_buf(idx)  <= sfd_buf(idx+1);
                            end loop;
                            sfd_buf(31)     <= tri_in;
                        end if;                
                        rx_counter_bits_v  := 0;
                        rx_counter_bytes_v := 0;
                        rx_data_len_v   := 0;                
                        data_from_cc1000r <= (others => '0');
                    when DATA =>
                        if (rx_counter_bits_v=0) then 
                            data_from_cc1000r <= (others => '0');
                        end if;                
                        data_from_cc1000r(rx_counter_bits_v) <= tri_in;
                        adc_request_now                  <= '0';  -- zdejmij zadanie dla przetwornika    
                        -- get number of bytes of data from the first received byte:
                        if (rx_counter_bits_v = 8 and rx_counter_bytes_v = 1) then
                            rx_data_len_v := conv_integer(data_from_cc1000r(7 downto 0));  -- get the count of data
                        end if;
                        
                        -- end of data:
                        if (rx_counter_bytes_v >= 1 and rx_counter_bytes_v >= rx_data_len_v) then -- data_len_v is valid when counter_bytes_v >= 1 
                            rx_state    <= WAITING;
                            rx_busy     <= '0';
                            rx_done     <= '1';        
                            data_from_cc1000r(rx_counter_bits_v) <= '0'; -- discard the last bit that came too late          
                            rx_fifo_write_v := TRUE;
                        end if;

                        -- check data_length, if too big then restart:
                        if rx_counter_bits_v = 8 and rx_counter_bytes_v = 1 and rx_data_len_v > DATA_BYTES_MAX then
                            rx_state    <= PREAMBLE;
                            rx_busy     <= '0';
                            rx_done     <= '0';
                            rx_counter_bits_v  := 0;
                            rx_counter_bytes_v := 0;
                        end if;
                                                                       
                        rx_counter_bits_v := rx_counter_bits_v + 1;
                        
                        if (rx_counter_bits_v = 8 or rx_counter_bits_v = 16 or rx_counter_bits_v = 24 or rx_counter_bits_v = 32) then
                            rx_counter_bytes_v := rx_counter_bytes_v + 1;
                        end if;
                        
                        if (rx_counter_bits_v >= 32) then
                            rx_fifo_write_v := TRUE;
                        end if;
                        
                        if (rx_fifo_write_v) then
                            rx_fifo_wr    <= '1';
                            if rx_fifo_full = '1' then
                                rx_state    <= PREAMBLE;                
                                rx_busy     <= '0';
                                rx_done     <= '0';
                                rx_counter_bytes_v := 0;
                                rx_fifo_wr    <= '0';
                            end if;
                            rx_counter_bits_v    := 0;
                        end if;                                                                                               
                    when WAITING =>
                        null;
                end case;
            end if;
        end if;
        dbg_o(81 downto 76) <= conv_std_logic_vector(rx_counter_bits_v, 6);
        dbg_o(87 downto 82) <= conv_std_logic_vector(rx_counter_bytes_v, 6);
        dbg_o(93 downto 88) <= conv_std_logic_vector(rx_data_len_v, 6);        
    end if;    
  end process RX_PROC;  
  
end architecture;
